/*
 * Copyright 2008-2009 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package javarequirementstracer.mojo;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Properties;
import java.util.Set;

import javarequirementstracer.FileUtils;
import javarequirementstracer.JavaRequirementsTracer;
import javarequirementstracer.JavaRequirementsTracerBean;
import javarequirementstracer.Requirements;

import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.factory.ArtifactFactory;
import org.apache.maven.artifact.repository.ArtifactRepository;
import org.apache.maven.artifact.resolver.ArtifactNotFoundException;
import org.apache.maven.artifact.resolver.ArtifactResolutionException;
import org.apache.maven.artifact.resolver.ArtifactResolver;
import org.apache.maven.artifact.resolver.filter.ScopeArtifactFilter;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.project.MavenProject;
import org.apache.maven.project.MavenProjectBuilder;
import org.apache.maven.project.ProjectBuildingException;
import org.apache.maven.project.artifact.InvalidDependencyVersionException;

/**
 * @author Ruud de Jong
 * @goal trace
 * @phase test-compile
 * @requiresDependencyResolution test
 */
public class JrtMojo extends AbstractMojo {

	private static final String REPORTER_PROPERTIES_FILENAME = "/reporter.properties";
	private static final String REPORTER_GROUPID = "reporter.groupid";
	private static final String REPORTER_ARTIFACTID = "reporter.artifactid";
	private static final String REPORTER_VERSION = "reporter.version";
	private static final String REPORTER_SCOPE = "runtime";
	private static final String REPORTER_TYPE = "jar";


	/**
	 * Comma separated list of properties filenames each containing all tuning parameters for a report.
	 * For each a report will be generated.
	 * Default: {@value JavaRequirementsTracer#DEFAULT_PARAMS_FILENAMES}.
	 * If a file does not exist the plugin will skip generating the report and print a log message. 
	 * 
	 * @parameter expression="${traceParamsFilenames}"
	 */
	private String traceParamsFilenames; 

	/**
	 * The build number to display in the report.
	 * 
	 * @parameter expression="${buildNumber}"
	 */
	private String buildNumber;

	/**
	 * The Maven Project Object
	 *
	 * @parameter expression="${project}"
	 * @required
	 * @readonly
	 */
	private MavenProject project;

	/**
	 * The classpath elements of the project being tested.
	 *
	 * @parameter expression="${project.testClasspathElements}"
	 * @required
	 * @readonly
	 */
	private List<String> testClasspathElements;

	/**
	 * ArtifactRepository of the localRepository. To obtain the directory of localRepository in unit tests use
	 * System.setProperty( "localRepository").
	 *
	 * @parameter expression="${localRepository}"
	 * @required
	 * @readonly
	 */
	private ArtifactRepository localRepository;

	/**
	 * ArtifactFactory component.
	 *
	 * @component
	 */
	private ArtifactFactory artifactFactory;

    /**
     * Project builder -- builds a model from a pom.xml
     *
     * @component role="org.apache.maven.project.MavenProjectBuilder"
     * @required
     * @readonly
     */
    private MavenProjectBuilder mavenProjectBuilder;

    /**
     * List of Remote Repositories used by the resolver
     *
     * @parameter expression="${project.remoteArtifactRepositories}"
     * @readonly
     * @required
     */
    private java.util.List remoteRepos;

    /**
     * Used to look up Artifacts in the remote repository.
     *
     * @component
     */
    protected ArtifactResolver artifactResolver;

	/**
	 * Maven plugin execute method.
	 * This will check whether the reporter should be run,
	 * assembles the classloader and start the reporter from the alternate classloader.
	 * @throws MojoExecutionException when an error occurs.
	 */
    @Requirements({"UC-Goal-Trace", "UC-No-Spring"})
	public final void execute() throws MojoExecutionException {
		info("Project basedir = " + this.project.getBasedir());
		info("Build number = " + this.buildNumber);
		info("Trace params filenames = " + this.traceParamsFilenames);

		if (this.traceParamsFilenames == null) {
			this.traceParamsFilenames = JavaRequirementsTracer.DEFAULT_PARAMS_FILENAMES;
		}
		
		List<URL> classpathUrlList = new ArrayList<URL>();
		classpathUrlList.addAll(getReporterArtifactUrls());
		classpathUrlList.addAll(getTestClasspathUrls());

		URLClassLoader classLoader = new URLClassLoader(classpathUrlList.toArray(new URL[classpathUrlList.size()]));

		runTracerWithClassloader(classLoader);
	}

	/**
	 * Runs the Java Requirements Tracer.
	 * This tracer needs to be loaded by the alternative classloader, which
	 * has all the right classpaths and dependencies added.
	 * This can ONLY be done through reflection, because the same class from different
	 * classloaders will ALWAYS be not-equal AND will ALWAYS give class cast exceptions.
	 * @param classLoader the Class Loader to use.
	 * @throws MojoExecutionException when shit hits the fan.
	 */
	private void runTracerWithClassloader(ClassLoader classLoader) throws MojoExecutionException {
		try {
			Class<?> tracerClass = classLoader.loadClass(JavaRequirementsTracerBean.class.getName());
			Object tracerInstance = tracerClass.getConstructor().newInstance();
			Method setClassLoaderMethod = tracerClass.getDeclaredMethod("setClassLoader", ClassLoader.class);
			setClassLoaderMethod.invoke(tracerInstance, classLoader);
			invokeStringSetter(tracerInstance, "setBuildNumber", this.buildNumber);
			String[] parFilenames = JavaRequirementsTracer.split(this.traceParamsFilenames);		
			for (String parFilename : parFilenames) {
				if (!FileUtils.isAbsolute(parFilename)) {
					parFilename = this.project.getBasedir() + "/" + parFilename;
				}
				File paramFile = new File(parFilename);
				if (!paramFile.exists()) {
					info("Skipping " + paramFile.getAbsolutePath() + " (file not found).");
					continue;
				}
				invokeStringSetter(tracerInstance, "setParamsFilename", parFilename);
				Method runMethod = tracerClass.getDeclaredMethod("run");
				runMethod.invoke(tracerInstance);
			}
		} catch (ClassNotFoundException e) {
			throw new MojoExecutionException("Could not load " + JavaRequirementsTracerBean.class.getName() + " from the alternate classloader.", e);
		} catch (NoSuchMethodException e) {
			throw new MojoExecutionException("Could not find constructor of " + JavaRequirementsTracerBean.class.getName() + ".", e);
		} catch (InvocationTargetException e) {
			throw new MojoExecutionException("Could not find invoke constructor of " + JavaRequirementsTracerBean.class.getName() + ".", e);
		} catch (IllegalAccessException e) {
			throw new MojoExecutionException("Could not access constructor of " + JavaRequirementsTracerBean.class.getName() + ".", e);
		} catch (InstantiationException e) {
			throw new MojoExecutionException("Could not create instance of " + JavaRequirementsTracerBean.class.getName() + ".", e);
		}
	}

	private void invokeStringSetter(Object instance, String methodName, String value) throws NoSuchMethodException, InvocationTargetException, IllegalAccessException {
		if (value != null) {
			Method stringSetterMethod = instance.getClass().getDeclaredMethod(methodName, String.class);
			stringSetterMethod.invoke(instance, value);
		}
	}

	/**
	 * Convert the test class path elements into URLs.
	 * @return a list of URLs.
	 * @throws MojoExecutionException when URLs are invalid.
	 */
	private List<URL> getTestClasspathUrls() throws MojoExecutionException {
		List<URL> urlList = new ArrayList<URL>();
		for (String testClasspath : this.testClasspathElements) {
			URL url = pathToUrl(testClasspath);
			urlList.add(url);
		}
		return urlList;
	}

	/**
	 * Convert a String to a File to an URL.
	 * @param path the string-form of a file or path reference to convert.
	 * @return the URL.
	 * @throws MojoExecutionException then the URL is invalid.
	 */
	private URL pathToUrl(String path) throws MojoExecutionException{
		try {
			File file = new File(path);
			return file.toURI().toURL();
		} catch (MalformedURLException e) {
			throw new MojoExecutionException("Invalid URL: " + path, e);
		}
	}

	/**
	 * Retrieve the URLs of the reporter artifact.
     * This returns a list of URLs to the reporter artifact and it's dependencies.
	 * This reads the properties file from the JAR of this MOJO and consult the artifact factory for the URL.
	 * @return the reporter URLs.
	 * @throws MojoExecutionException when there are problems with the properties or the URL.
	 */
	private List<URL> getReporterArtifactUrls() throws MojoExecutionException {
        Artifact reporterArtifact = getReporterArtifact();
        Set<Artifact> dependenciesArtifacts = resolveDependencyArtifacts(reporterArtifact);

        Set<Artifact> allArtifacts = new HashSet<Artifact>();
        allArtifacts.add(reporterArtifact);
        allArtifacts.addAll(dependenciesArtifacts);

        return convertArtifactsIntoUrls(allArtifacts);
	}

    /**
     * Convert Artifacts into URLs
     * @param artifactSet A set of artifacts
     * @return the list of URLs
     * @throws MojoExecutionException
     */
    private List<URL> convertArtifactsIntoUrls(Set<Artifact> artifactSet) throws MojoExecutionException {
        List<URL> urlList = new ArrayList<URL>();
        for (Artifact artifact : artifactSet) {
            URL url = pathToUrl(this.localRepository.getBasedir() + "/" + this.localRepository.pathOf(artifact));
            urlList.add(url);
            info("Reporter URL = " + url.toString());
        }
		return urlList;

    }

    /**
     * Construct the reporter artifact, based on the Mojo properties.
     * @return the reporter artifact
     * @throws MojoExecutionException
     */
    private Artifact getReporterArtifact() throws MojoExecutionException {
        InputStream propertiesStream = JrtMojo.class.getResourceAsStream(REPORTER_PROPERTIES_FILENAME);
        if (propertiesStream == null) {
            throw new MojoExecutionException("Cannot find properties file: " + REPORTER_PROPERTIES_FILENAME);
        }
        Properties reporterProperties = new Properties();
        try {
            reporterProperties.load(propertiesStream);
        } catch (IOException e) {
            throw new MojoExecutionException("Cannot load properties from file: " + REPORTER_PROPERTIES_FILENAME, e);
        }
        return this.artifactFactory.createArtifact(
                reporterProperties.getProperty(REPORTER_GROUPID),
                reporterProperties.getProperty(REPORTER_ARTIFACTID),
                reporterProperties.getProperty(REPORTER_VERSION), REPORTER_SCOPE, REPORTER_TYPE);
    }

    /**
     * This method resolves the dependency artifacts from the given artifact.
     *
     * @param artifact The Artifact.
     * @return resolved set of dependency artifacts.
     *
     * @throws MojoExecutionException
     */
    private Set<Artifact> resolveDependencyArtifacts(Artifact artifact) throws MojoExecutionException
    {
        Set<Artifact> nonOptionalDependencies = new HashSet<Artifact>();
        try {
            MavenProject project = mavenProjectBuilder.buildFromRepository(artifact, this.remoteRepos, this.localRepository);
            Set<Artifact> dependencyArtifacts = project.createArtifacts( this.artifactFactory, Artifact.SCOPE_RUNTIME,
                                                        new ScopeArtifactFilter( Artifact.SCOPE_RUNTIME ) );
            Set<Artifact> transitiveDependencies = new HashSet<Artifact>();
            for (Artifact dependencyArtifact : dependencyArtifacts) {
                if (!dependencyArtifact.isOptional()) {
                    // resolve the new artifact
                    this.artifactResolver.resolve( dependencyArtifact, this.remoteRepos, this.localRepository );
                    nonOptionalDependencies.add(dependencyArtifact);
                    transitiveDependencies.addAll(resolveDependencyArtifacts(dependencyArtifact));
                }
            }
            nonOptionalDependencies.addAll(transitiveDependencies);
            return nonOptionalDependencies;
        } catch (ProjectBuildingException e) {
            throw new MojoExecutionException("Could not resolve dependencies for reporter artifact.", e);
        } catch (InvalidDependencyVersionException e) {
            throw new MojoExecutionException("Invalid dependency version.", e);
        } catch (ArtifactResolutionException e) {
            throw new MojoExecutionException("Problem resolving artifact.", e);
        } catch (ArtifactNotFoundException e) {
            throw new MojoExecutionException("Artifact not found.", e);
        }
    }
    
    private void info(String str) {
		String prefix = "JRT-Trace: ";
		getLog().info(prefix + str);
    }
}